-- Source code from
--   Purely Functional Data Structures
--   Chris Okasaki
--   Cambridge University Press, 1998
--
-- Copyright (c) 1998 Cambridge University Press

module BankersDeque (module Deque, BankersDeque, showBD) where
import Prelude hiding (head,tail,last,init)
import Data.List (intercalate)
import NestedShowable (NestedShowable (showNested))
import Deque

c :: Int
c = 3

data BankersDeque a = BD Int [a] Int [a] deriving (Show)

showBD :: (a -> String) -> BankersDeque a -> String
showBD f (BD xsz x ysz y) =
    show xsz ++ " " ++ "(" ++ intercalate "," (map f x) ++ ") " ++
    show ysz ++ " " ++ "(" ++ intercalate "," (map f y) ++ "))"

instance NestedShowable BankersDeque where
    showNested = showBD

check :: Int -> [a] -> Int -> [a] -> BankersDeque a
check lenf f lenr r
    | lenf > c*lenr + 1 =
      let i = (lenf+lenr) `div` 2
          j = lenf+lenr-i
          f' = take i f
          r' = r ++ reverse (drop i f)
      in BD i f' j r'
    | lenr > c*lenf + 1 =
      let j = (lenf+lenr) `div` 2
          i = lenf+lenr-j
          r' = take j r
          f' = f ++ reverse (drop j r)
      in BD i f' j r'
    | otherwise = BD lenf f lenr r

instance Deque BankersDeque where
  empty = BD 0 [] 0 []
  isEmpty (BD lenf _ lenr _) = lenf+lenr == 0

  cons x (BD lenf f lenr r) = check (lenf+1) (x:f) lenr r

  head (BD _    []    _    _) = error "BankersDeque.head: empty deque"
  head (BD _    (x:_) _    _) = x

  tail (BD _    []     _    _) = error "BankersDeque.tail: empty deque"
  tail (BD lenf (_:f') lenr r) = check (lenf-1) f' lenr r

  snoc (BD lenf f lenr r) x = check lenf f (lenr+1) (x:r)

  last (BD _    _ _     [])    = error "BankersDeque.last: empty deque"
  last (BD _    _ _     (x:_)) = x

  init (BD _    _ _    [])     = error "BankersDeque.init: empty deque"
  init (BD lenf f lenr (_:r')) = check lenf f (lenr-1) r'
